import logging
from typing import Dict, Any, Tuple

import torch as to
from torch import nn
from torch.nn.modules.module import Module
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader

from message_passing_nn.data.preprocessor import Preprocessor
from message_passing_nn.utils.loss_function_selector import LossFunctionSelector
from message_passing_nn.utils.optimizer_selector import OptimizerSelector


class ModelTrainer:
    def __init__(self, model: nn.Module, preprocessor: Preprocessor, device: str, normalize: bool = False) -> None:
        self.model = model
        self.preprocessor = preprocessor
        self.loss_function = None
        self.optimizer = None
        self.device = device
        self.normalize = normalize

    def instantiate_attributes(self,
                               data_dimensions: Tuple,
                               configuration_dictionary: Dict) -> None:
        node_features_size, adjacency_matrix_size, labels_size = data_dimensions
        number_of_nodes = adjacency_matrix_size[0]
        number_of_node_features = node_features_size[1]
        fully_connected_layer_output_size = labels_size[0]
        self.model = self.model.of(time_steps=configuration_dictionary['time_steps'],
                                   number_of_nodes=number_of_nodes,
                                   number_of_node_features=number_of_node_features,
                                   fully_connected_layer_input_size=number_of_nodes * number_of_node_features,
                                   fully_connected_layer_output_size=fully_connected_layer_output_size,
                                   device=self.device)
        self.model.to(self.device)
        self.loss_function = self._instantiate_the_loss_function(
            LossFunctionSelector.load_loss_function(configuration_dictionary['loss_function']))
        self.optimizer = self._instantiate_the_optimizer(
            OptimizerSelector.load_optimizer(configuration_dictionary['optimizer']))

    def do_train(self, training_data: DataLoader, epoch: int) -> float:
        training_loss = 0.0
        for features, labels in training_data:
            node_features, adjacency_matrix = features
            node_features, adjacency_matrix, labels = node_features.to(self.device), \
                                                      adjacency_matrix.to(self.device), \
                                                      labels.to(self.device)
            current_batch_size = self._get_current_batch_size(labels)
            if self.normalize:
                node_features = self.preprocessor.normalize(node_features)
                labels = self.preprocessor.normalize(labels)
            self.optimizer.zero_grad()
            outputs = self.model.forward(node_features,
                                         adjacency_matrix=adjacency_matrix,
                                         batch_size=current_batch_size)
            loss = self.loss_function(outputs, labels)
            training_loss += self._do_backpropagate(loss, training_loss)
        training_loss /= len(training_data)
        self.get_logger().info('[Iteration %d] training loss: %.3f' % (epoch, training_loss))
        return training_loss

    def do_evaluate(self, evaluation_data: DataLoader, epoch: int = None) -> float:
        with to.no_grad():
            evaluation_loss = 0.0
            if len(evaluation_data):
                for features_validation, labels_validation in evaluation_data:
                    node_features, adjacency_matrix = features_validation
                    node_features, adjacency_matrix, labels_validation = node_features.to(self.device), \
                                                                         adjacency_matrix.to(self.device), \
                                                                         labels_validation.to(self.device)
                    if self.normalize:
                        node_features = self.preprocessor.normalize(node_features)
                        labels_validation = self.preprocessor.normalize(labels_validation)
                    current_batch_size = self._get_current_batch_size(labels_validation)
                    outputs = self.model.forward(node_features, adjacency_matrix, current_batch_size)
                    loss = self.loss_function(outputs, labels_validation)
                    evaluation_loss += float(loss)
                evaluation_loss /= len(evaluation_data)
                if epoch is not None:
                    self.get_logger().info('[Iteration %d] validation loss: %.3f' % (epoch, evaluation_loss))
                else:
                    self.get_logger().info('Test loss: %.3f' % evaluation_loss)
            else:
                self.get_logger().warning('No evaluation data found!')
        return evaluation_loss

    def _do_backpropagate(self, loss: to.Tensor, training_loss: float) -> float:
        loss.backward()
        self.optimizer.step()
        training_loss += loss.item()
        return training_loss

    @staticmethod
    def _instantiate_the_loss_function(loss_function: Module) -> Module:
        return loss_function()

    def _instantiate_the_optimizer(self, optimizer: Any) -> Optimizer:
        model_parameters = list(self.model.parameters())
        return optimizer(model_parameters, lr=0.001, momentum=0.9)

    @staticmethod
    def _get_current_batch_size(features: to.Tensor) -> int:
        return len(features)

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



from typing import List, Tuple

import torch as to
from torch.utils.data import Dataset


class GraphDataset(Dataset):
    def __init__(self,
                 dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                 normalize_features: bool = True,
                 normalize_labels: bool = True) -> None:
        self.features = self._to_list(dataset)
        self.labels = self._extract_labels(dataset)
        self.normalize_features = normalize_features
        self.normalize_labels = normalize_labels

    def __len__(self) -> int:
        return len(self.features)

    def __getitem__(self, index: int) -> Tuple:
        return self.features[index], self.labels[index]

    @staticmethod
    def _to_list(dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]]) -> List[Tuple[to.Tensor, to.Tensor]]:
        return [(dataset[index][0], dataset[index][1]) for index in range(len(dataset))]

    @staticmethod
    def _extract_labels(dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]]) -> List[to.Tensor]:
        return [dataset[index][2] for index in range(len(dataset))]




from unittest import TestCase

import torch as to

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestGraphPreprocessor(TestCase):
    def setUp(self) -> None:
        self.data_preprocessor = DataPreprocessor()

    def test_train_validation_test_split(self):
        # Given
        dataset_length = 10
        features = BASE_GRAPH_NODE_FEATURES
        adjacency_matrix = BASE_GRAPH
        labels = BASE_GRAPH.view(-1)
        raw_dataset = [(features, adjacency_matrix, labels) for i in range(dataset_length)]
        train_validation_test_split_expected = [7, 2, 1]

        # When
        train_validation_test_split = self.data_preprocessor.train_validation_test_split(raw_dataset,
                                                                                         batch_size=1,
                                                                                         maximum_number_of_features=-1,
                                                                                         maximum_number_of_nodes=-1,
                                                                                         validation_split=0.2,
                                                                                         test_split=0.1)
        train_validation_test_split = [len(dataset) for dataset in train_validation_test_split]

        # Then
        self.assertEqual(train_validation_test_split_expected, train_validation_test_split)

    def test_equalize_sizes(self):
        # Given
        node_features_1 = to.ones((2, 5))
        node_features_2 = to.ones((3, 5))
        adjacency_matrix_1 = to.ones(2, 2)
        adjacency_matrix_2 = to.ones(3, 3)
        labels_1 = to.ones(15)
        labels_2 = to.ones(20)
        raw_dataset = [(node_features_1, adjacency_matrix_1, labels_1),
                       (node_features_2, adjacency_matrix_2, labels_2)]

        # When
        equalized_dataset = self.data_preprocessor.equalize_dataset_dimensions(raw_dataset, -1, -1)
        node_features_1, adjacency_matrix_1, labels_1 = equalized_dataset[0]

        # Then
        self.assertEqual(node_features_1.size(), node_features_2.size())
        self.assertEqual(adjacency_matrix_1.size(), adjacency_matrix_2.size())
        self.assertEqual(labels_1.size(), labels_2.size())

    def test_extract_data_dimensions(self):
        # Given
        dataset_length = 1
        features = BASE_GRAPH_NODE_FEATURES
        adjacency_matrix = BASE_GRAPH
        labels = BASE_GRAPH.view(-1)
        raw_dataset = [(features, adjacency_matrix, labels) for i in range(dataset_length)]
        data_dimensions_expected = (features.size(), adjacency_matrix.size(), labels.size())

        # When
        data_dimensions = self.data_preprocessor.extract_data_dimensions(raw_dataset)

        # Then
        self.assertEqual(data_dimensions_expected, data_dimensions)

    def test_flatten_when_sizes_match(self):
        # Given
        dataset_length = 2
        labels = BASE_GRAPH.view(-1)
        tensors = to.cat((labels, labels))
        tensors_flattened_expected = tensors.view(-1)

        # When
        tensors_flattened = self.data_preprocessor.flatten(tensors, desired_size=dataset_length * len(labels))

        # Then
        self.assertTrue(to.allclose(tensors_flattened_expected, tensors_flattened))

    def test_flatten_when_sizes_do_not_match(self):
        # Given
        dataset_length = 3
        labels = BASE_GRAPH.view(-1)
        tensors = to.cat((labels, labels))
        tensors_flattened_expected = to.cat((tensors.view(-1), to.zeros_like(labels)))

        # When
        tensors_flattened = self.data_preprocessor.flatten(tensors, desired_size=dataset_length * len(labels))

        # Then
        self.assertTrue(to.allclose(tensors_flattened_expected, tensors_flattened))



import numpy as np
import torch as to

NOT_SYMMETRIC_GRAPH = to.tensor([[0, 1, 1, 1],
                                 [1, 0, 1, 0],
                                 [1, 1, 0, 1],
                                 [0, 0, 1, 0]]).float()

BASE_GRAPH = to.tensor([[0, 1, 1, 0],
                        [1, 0, 1, 0],
                        [1, 1, 0, 1],
                        [0, 0, 1, 0]]).float()
BASE_GRAPH_NODE_FEATURES = to.tensor([[1, 2], [1, 1], [2, 0.5], [0.5, 0.5]]).float()
BASE_UNITY_MATRIX = np.ones((BASE_GRAPH_NODE_FEATURES.shape[1], BASE_GRAPH_NODE_FEATURES.shape[1]))
BASE_UNITY_VECTOR = np.ones((BASE_GRAPH_NODE_FEATURES.shape[1]))
BASE_ZEROS_MATRIX = np.zeros((BASE_GRAPH_NODE_FEATURES.shape[1], BASE_GRAPH_NODE_FEATURES.shape[1]))

BASE_GRAPH_EDGE_FEATURES = to.tensor([[[0.0, 0.0],   [1.0, 2.0], [2.0, 0.5],   [0.0, 0.0]],
                                      [[1.0, 2.0],   [0.0, 0.0], [1.0, 1.0],   [0.0, 0.0]],
                                      [[2.0, 0.5],   [1.0, 1.0], [0.0, 0.0],   [0.5, 0.5]],
                                      [[0.0, 0.0],   [0.0, 0.0], [0.5, 0.5],   [0.0, 0.0]]]).float()
BASE_W_MATRIX = to.tensor([[BASE_ZEROS_MATRIX, BASE_UNITY_MATRIX, BASE_UNITY_MATRIX, BASE_ZEROS_MATRIX],
                           [BASE_UNITY_MATRIX, BASE_ZEROS_MATRIX, BASE_UNITY_MATRIX, BASE_ZEROS_MATRIX],
                           [BASE_UNITY_MATRIX, BASE_UNITY_MATRIX, BASE_ZEROS_MATRIX, BASE_UNITY_MATRIX],
                           [BASE_ZEROS_MATRIX, BASE_ZEROS_MATRIX, BASE_UNITY_MATRIX, BASE_ZEROS_MATRIX]]).float()
BASE_U_MATRIX = to.tensor([BASE_UNITY_MATRIX,
                           BASE_UNITY_MATRIX,
                           BASE_UNITY_MATRIX,
                           BASE_UNITY_MATRIX]).float()

BASE_B_VECTOR = to.tensor(BASE_UNITY_VECTOR).float()

MULTIPLICATION_FACTOR = 0.1



from abc import ABCMeta, abstractmethod
from typing import List, Tuple, Any


class Repository(metaclass=ABCMeta):
    def __init__(self):
        pass

    @abstractmethod
    def save(self, filename: str, dataset_name: str) -> None:
        pass

    @abstractmethod
    def get_all_data(self) -> List[Tuple[Any, Any, Any]]:
        pass



from typing import Tuple, List

import torch as to


class Node:
    def __init__(self, all_node_features: to.Tensor, adjacency_matrix: to.Tensor, node_id: int):
        self.node_id = int(node_id)
        self.features = self._get_features_of_specific_node(all_node_features)
        self.symmetry = self.is_adjacency_matrix_symmetric(adjacency_matrix)
        self.all_neighbors = self._get_all_neighbors(adjacency_matrix)
        self.neighbors = self._get_neighbors(adjacency_matrix)
        self.neighbors_count = len(self.neighbors)

    @staticmethod
    def is_adjacency_matrix_symmetric(adjacency_matrix: to.Tensor) -> bool:
        return to.allclose(adjacency_matrix, adjacency_matrix.t())

    def _get_features_of_specific_node(self, all_node_features: to.Tensor) -> to.Tensor:
        return all_node_features[self.node_id]

    def _get_all_neighbors(self, adjacency_matrix: to.Tensor) -> to.Tensor:
        return to.nonzero(adjacency_matrix[self.node_id], as_tuple=True)[0]

    def _get_neighbors(self, adjacency_matrix: to.Tensor) -> List:
        neighbors = self._get_all_neighbors(adjacency_matrix)
        if self.symmetry:
            neighbors = self._get_greater_neighbors(neighbors)
        return neighbors

    def _get_greater_neighbors(self, neighbors: List) -> List:
        return [neighbor for neighbor in neighbors if neighbor > self.node_id]

    def get_start_node_neighbors_without_end_node(self, end_node_id: int) -> Tuple:
        return self._remove_end_node_from_start_node_neighbors(end_node_id), [self.node_id]

    def _remove_end_node_from_start_node_neighbors(self, end_node_id: int) -> to.tensor:
        end_node_index = (self.all_neighbors == end_node_id).nonzero()[0][0].item()
        return to.cat((self.all_neighbors[:end_node_index],
                       self.all_neighbors[end_node_index + 1:])).tolist()


import os
from typing import List
from unittest import TestCase

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.graph_encoder import GraphEncoder
from message_passing_nn.repository.file_system_repository import FileSystemRepository
from message_passing_nn.trainer.model_trainer import ModelTrainer
from message_passing_nn.usecase.grid_search import GridSearch
from message_passing_nn.utils.saver import Saver
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestTraining(TestCase):
    def setUp(self) -> None:
        self.features = BASE_GRAPH_NODE_FEATURES
        self.adjacency_matrix = BASE_GRAPH
        self.labels = BASE_GRAPH.view(-1)
        self.dataset = 'training-test-data'
        self.tests_data_directory = 'tests/test_data/'
        tests_model_directory = 'tests/model_checkpoints'
        tests_results_directory = 'tests/grid_search_results'
        device = "cpu"
        self.repository = FileSystemRepository(self.tests_data_directory, self.dataset)
        self.data_preprocessor = DataPreprocessor()
        self.model_trainer = ModelTrainer(GraphEncoder, self.data_preprocessor, device)
        self.saver = Saver(tests_model_directory, tests_results_directory)

    def test_start_for_multiple_batches_of_the_same_size(self):
        # Given
        dataset_size = 6
        grid_search_dictionary = {
            "epochs": [10],
            "batch_size": [3],
            "maximum_number_of_nodes": [-1],
            "maximum_number_of_features": [-1],
            "validation_split": [0.2],
            "test_split": [0.1],
            "loss_function": ["MSE"],
            "optimizer": ["SGD"],
            "time_steps": [1],
            "validation_period": [5]
        }
        grid_search = GridSearch(self.repository,
                                 self.data_preprocessor,
                                 self.model_trainer,
                                 grid_search_dictionary,
                                 self.saver)

        features_filenames = [str(i) + '_training_features' + '.pickle' for i in range(dataset_size)]
        adjacency_matrix_filenames = [str(i) + '_training_adjacency-matrix' '.pickle' for i in range(dataset_size)]
        labels_filenames = [str(i) + '_training_labels' '.pickle' for i in range(dataset_size)]
        for i in range(dataset_size):
            self.repository.save(features_filenames[i], self.features)
            self.repository.save(adjacency_matrix_filenames[i], self.adjacency_matrix)
            self.repository.save(labels_filenames[i], self.labels)

        # When
        losses = grid_search.start()
        configuration_id = list(losses["training_loss"].keys())[0]

        # Then
        self.assertTrue(losses["training_loss"][configuration_id][grid_search_dictionary["epochs"][0]] > 0.0)
        self.assertTrue(
            losses["validation_loss"][configuration_id][grid_search_dictionary["validation_period"][0]] > 0.0)
        self.assertTrue(losses["test_loss"][configuration_id]["final_epoch"] > 0.0)

        # Tear down
        self._remove_files(dataset_size, features_filenames, adjacency_matrix_filenames, labels_filenames)

    def test_start_for_multiple_batches_of_differing_size(self):
        # Given
        dataset_size = 5
        grid_search_dictionary = {
            "epochs": [10],
            "batch_size": [3],
            "maximum_number_of_nodes": [-1],
            "maximum_number_of_features": [-1],
            "validation_split": [0.2],
            "test_split": [0.1],
            "loss_function": ["MSE"],
            "optimizer": ["SGD"],
            "time_steps": [1],
            "validation_period": [5]
        }
        grid_search = GridSearch(self.repository,
                                 self.data_preprocessor,
                                 self.model_trainer,
                                 grid_search_dictionary,
                                 self.saver)

        features_filenames = [str(i) + '_training_features' + '.pickle' for i in range(dataset_size)]
        adjacency_matrix_filenames = [str(i) + '_training_adjacency-matrix' '.pickle' for i in range(dataset_size)]
        labels_filenames = [str(i) + '_training_labels' '.pickle' for i in range(dataset_size)]
        for i in range(dataset_size):
            self.repository.save(features_filenames[i], self.features)
            self.repository.save(adjacency_matrix_filenames[i], self.adjacency_matrix)
            self.repository.save(labels_filenames[i], self.labels)

        # When
        losses = grid_search.start()
        configuration_id = list(losses["training_loss"].keys())[0]

        # Then
        self.assertTrue(losses["training_loss"][configuration_id][grid_search_dictionary["epochs"][0]] > 0.0)
        self.assertTrue(
            losses["validation_loss"][configuration_id][grid_search_dictionary["validation_period"][0]] > 0.0)
        self.assertTrue(losses["test_loss"][configuration_id]["final_epoch"] > 0.0)

        # Tear down
        self._remove_files(dataset_size, features_filenames, adjacency_matrix_filenames, labels_filenames)

    def test_start_a_grid_search(self):
        # Given
        dataset_size = 6
        grid_search_dictionary = {
            "epochs": [10, 15],
            "batch_size": [3, 4],
            "maximum_number_of_nodes": [-1],
            "maximum_number_of_features": [-1],
            "validation_split": [0.2],
            "test_split": [0.1],
            "loss_function": ["MSE"],
            "optimizer": ["SGD"],
            "time_steps": [1],
            "validation_period": [5]
        }
        grid_search = GridSearch(self.repository,
                                 self.data_preprocessor,
                                 self.model_trainer,
                                 grid_search_dictionary,
                                 self.saver)

        features_filenames = [str(i) + '_training_features' + '.pickle' for i in range(dataset_size)]
        adjacency_matrix_filenames = [str(i) + '_training_adjacency-matrix' '.pickle' for i in range(dataset_size)]
        labels_filenames = [str(i) + '_training_labels' '.pickle' for i in range(dataset_size)]
        for i in range(dataset_size):
            self.repository.save(features_filenames[i], self.features)
            self.repository.save(adjacency_matrix_filenames[i], self.adjacency_matrix)
            self.repository.save(labels_filenames[i], self.labels)

        # When
        losses = grid_search.start()
        configuration_id = list(losses["training_loss"].keys())[0]

        # Then
        self.assertTrue(losses["training_loss"][configuration_id][grid_search_dictionary["epochs"][0]] > 0.0)
        self.assertTrue(
            losses["validation_loss"][configuration_id][grid_search_dictionary["validation_period"][0]] > 0.0)
        self.assertTrue(losses["test_loss"][configuration_id]["final_epoch"] > 0.0)

        # Tear down
        self._remove_files(dataset_size, features_filenames, adjacency_matrix_filenames, labels_filenames)

    def _remove_files(self,
                      dataset_size: int,
                      features_filenames: List[str],
                      adjacency_matrix_filenames: List[str],
                      labels_filenames: List[str]) -> None:
        for i in range(dataset_size):
            os.remove(self.tests_data_directory + self.dataset + "/" + features_filenames[i])
            os.remove(self.tests_data_directory + self.dataset + "/" + adjacency_matrix_filenames[i])
            os.remove(self.tests_data_directory + self.dataset + "/" + labels_filenames[i])



import logging

import click
import sys

from message_passing_nn.create_message_passing_nn import create


@click.group("message-passing-nn")
@click.option('--debug', default=False, help='Set the logs to debug level', show_default=True, is_flag=True)
def main(debug):
    log_level = logging.DEBUG if debug else logging.INFO
    setup_logging(log_level)


@click.command('grid-search', help='Starts the grid search')
@click.argument('dataset_name', envvar='DATASET_NAME', type=str)
@click.argument('data_directory', envvar='DATA_DIRECTORY', type=str)
@click.argument('model_directory', envvar='MODEL_DIRECTORY', type=str)
@click.argument('results_directory', envvar='RESULTS_DIRECTORY', type=str)
@click.argument('device', envvar='DEVICE', type=str)
@click.argument('epochs', envvar='EPOCHS', type=str)
@click.argument('loss_function', envvar='LOSS_FUNCTION', type=str)
@click.argument('optimizer', envvar='OPTIMIZER', type=str)
@click.argument('batch_size', envvar='BATCH_SIZE', type=str)
@click.argument('maximum_number_of_features', envvar='MAXIMUM_NUMBER_OF_FEATURES', type=str)
@click.argument('maximum_number_of_nodes', envvar='MAXIMUM_NUMBER_OF_NODES', type=str)
@click.argument('validation_split', envvar='VALIDATION_SPLIT', type=str)
@click.argument('test_split', envvar='TEST_SPLIT', type=str)
@click.argument('time_steps', envvar='TIME_STEPS', type=str)
@click.argument('validation_period', envvar='VALIDATION_PERIOD', type=str)
def start_training(dataset_name: str,
                   data_directory: str,
                   model_directory: str,
                   results_directory: str,
                   device: str,
                   epochs: str,
                   loss_function: str,
                   optimizer: str,
                   batch_size: str,
                   maximum_number_of_features: str,
                   maximum_number_of_nodes: str,
                   validation_split: str,
                   test_split: str,
                   time_steps: str,
                   validation_period: str) -> None:
    get_logger().info("Starting training")
    message_passing_nn = create(dataset_name,
                                data_directory,
                                model_directory,
                                results_directory,
                                device,
                                epochs,
                                loss_function,
                                optimizer,
                                batch_size,
                                maximum_number_of_features,
                                maximum_number_of_nodes,
                                validation_split,
                                test_split,
                                time_steps,
                                validation_period)
    message_passing_nn.start()


def setup_logging(log_level):
    get_logger().setLevel(log_level)

    logOutputFormatter = logging.Formatter(
        '%(asctime)s %(levelname)s - %(message)s [%(filename)s:%(lineno)s] [%(relativeCreated)d]')

    stdoutStreamHandler = logging.StreamHandler(sys.stdout)
    stdoutStreamHandler.setLevel(log_level)
    stdoutStreamHandler.setFormatter(logOutputFormatter)

    get_logger().addHandler(stdoutStreamHandler)

    stderrStreamHandler = logging.StreamHandler(sys.stdout)
    stderrStreamHandler.setLevel(logging.WARNING)
    stderrStreamHandler.setFormatter(logOutputFormatter)

    get_logger().addHandler(stderrStreamHandler)


def get_logger() -> logging.Logger:
    return logging.getLogger('message_passing_nn')


main.add_command(start_training)

if __name__ == '__main__':
    main()



import logging
from typing import Tuple, List

import torch as to
from torch import nn
from torch.utils.data import DataLoader

from message_passing_nn.data.graph_dataset import GraphDataset
from message_passing_nn.data.preprocessor import Preprocessor


class DataPreprocessor(Preprocessor):
    def __init__(self):
        super().__init__()

    def train_validation_test_split(self,
                                    dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                                    batch_size: int,
                                    maximum_number_of_nodes: int,
                                    maximum_number_of_features: int,
                                    validation_split: float = 0.2,
                                    test_split: float = 0.1) -> Tuple[DataLoader, DataLoader, DataLoader]:
        test_index, validation_index = self._get_validation_and_test_indexes(dataset,
                                                                             validation_split,
                                                                             test_split)
        training_data = DataLoader(GraphDataset(dataset[:validation_index]), batch_size)
        if validation_split:
            validation_data = DataLoader(GraphDataset(dataset[validation_index:test_index]), batch_size)
        else:
            validation_data = DataLoader(GraphDataset([]))
        if test_split:
            test_data = DataLoader(GraphDataset(dataset[test_index:]), batch_size)
        else:
            test_data = DataLoader(GraphDataset([]))
        return training_data, validation_data, test_data

    def equalize_dataset_dimensions(self,
                                    dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                                    maximum_number_of_nodes: int,
                                    maximum_number_of_features: int) \
            -> List[Tuple[to.Tensor, to.Tensor, to.Tensor]]:

        adjacency_matrix_max_size, features_max_size, labels_max_size = self._get_maximum_data_size(dataset,
                                                                                                    maximum_number_of_nodes)

        preprocessed_dataset = self._equalize_sizes(dataset,
                                                    adjacency_matrix_max_size,
                                                    features_max_size,
                                                    labels_max_size,
                                                    maximum_number_of_nodes,
                                                    maximum_number_of_features)
        return preprocessed_dataset

    @staticmethod
    def extract_data_dimensions(dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]]) -> Tuple:
        node_features_size = dataset[0][0].size()
        adjacency_matrix_size = dataset[0][1].size()
        labels_size = dataset[0][2].size()
        return node_features_size, adjacency_matrix_size, labels_size

    @staticmethod
    def flatten(tensors: to.Tensor, desired_size: int = 0) -> to.Tensor:
        flattened_tensor = tensors.view(-1)
        if 0 < desired_size != len(flattened_tensor):
            flattened_tensor = DataPreprocessor._pad_zeros(flattened_tensor, desired_size)
        return flattened_tensor

    @staticmethod
    def normalize(tensor: to.Tensor) -> to.Tensor:
        if tensor.size()[0] > 1:
            normalizer = nn.BatchNorm1d(tensor.size()[1], affine=False)
            return normalizer(tensor)
        else:
            return tensor

    @staticmethod
    def _pad_zeros(flattened_tensor: to.Tensor, desired_size: int) -> to.Tensor:
        size_difference = abs(len(flattened_tensor) - desired_size)
        flattened_tensor = to.cat((flattened_tensor, to.zeros(size_difference)))
        return flattened_tensor

    @staticmethod
    def _get_validation_and_test_indexes(dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                                         validation_split: float,
                                         test_split: float) -> Tuple[int, int]:
        validation_index = int((1 - validation_split - test_split) * len(dataset))
        test_index = int((1 - test_split) * len(dataset))
        return test_index, validation_index

    def _equalize_sizes(self,
                        dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                        adjacency_matrix_max_size: List[int],
                        features_max_size: List[int],
                        labels_max_size: List[int],
                        maximum_number_of_nodes: int,
                        maximum_number_of_features: int) -> List[Tuple[to.Tensor, to.Tensor, to.Tensor]]:
        preprocessed_dataset = []
        for features, adjacency_matrix, labels in dataset:
            if self._is_graph_size_less_than_maximum_allowed(maximum_number_of_nodes, adjacency_matrix):
                continue
            if maximum_number_of_features < 0:
                maximum_number_of_features = features.size()[1]
            features_preprocessed = to.zeros((features_max_size[0], maximum_number_of_features))
            adjacency_matrix_preprocessed = to.zeros((adjacency_matrix_max_size[0], adjacency_matrix_max_size[1]))
            labels_preprocessed = to.zeros((labels_max_size[0]))

            features_preprocessed[:features.size()[0], :maximum_number_of_features] = features[:,
                                                                                      :maximum_number_of_features]
            adjacency_matrix_preprocessed[:adjacency_matrix.size()[0], :adjacency_matrix.size()[1]] = adjacency_matrix
            labels_preprocessed[:labels.size()[0]] = labels
            preprocessed_dataset.append((features_preprocessed, adjacency_matrix_preprocessed, labels_preprocessed))
        return preprocessed_dataset

    def _get_maximum_data_size(self,
                               dataset: List[Tuple[to.Tensor, to.Tensor, to.Tensor]],
                               maximum_number_of_nodes: int) \
            -> Tuple[List, List, List]:
        features_max_size = [0, 0]
        adjacency_matrix_max_size = [0, 0]
        labels_max_size = [0]
        for features, adjacency_matrix, labels in dataset:
            if self._is_graph_size_less_than_maximum_allowed(maximum_number_of_nodes, adjacency_matrix):
                continue
            if features.size()[0] > features_max_size[0]:
                features_max_size[0] = features.size()[0]
            if features.size()[1] > features_max_size[1]:
                features_max_size[1] = features.size()[1]
            if adjacency_matrix.size()[0] > adjacency_matrix_max_size[0]:
                adjacency_matrix_max_size = adjacency_matrix.size()
            if labels.size()[0] > labels_max_size[0]:
                labels_max_size = labels.size()
        return adjacency_matrix_max_size, features_max_size, labels_max_size

    @staticmethod
    def _is_graph_size_less_than_maximum_allowed(maximum_number_of_nodes: int, adjacency_matrix: to.Tensor) -> bool:
        return 0 < maximum_number_of_nodes < adjacency_matrix.size()[0]

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



import torch as to

optimizers = {
    "Adadelta": to.optim.Adadelta,
    "Adagrad": to.optim.Adagrad,
    "Adam": to.optim.Adam,
    "AdamW": to.optim.AdamW,
    "SparseAdam": to.optim.SparseAdam,
    "Adamax": to.optim.Adamax,
    "ASGD": to.optim.ASGD,
    "LBFGS": to.optim.LBFGS,
    "RMSprop": to.optim.RMSprop,
    "Rprop": to.optim.Rprop,
    "SGD": to.optim.SGD,
}



from unittest import TestCase

from message_passing_nn.utils.grid_search_parameters_parser import GridSearchParametersParser


class TestGridSearchParametersParser(TestCase):
    def test_get_grid_search_dictionary(self):
        # Given
        epochs = "10&15&5"
        loss_function_selection = "MSE&CrossEntropy"
        optimizer_selection = "SGD&Adam"
        batch_size = "5"
        maximum_number_of_nodes = "-1"
        maximum_number_of_features = "-1"
        validation_split = "0.2"
        test_split = "0.1&0.2&2"
        time_steps = "10"
        validation_period = "5"
        grid_search_dictionary_expected = {
            "epochs": [10, 11, 12, 13, 15],
            "loss_function": ["MSE", "CrossEntropy"],
            "optimizer": ["SGD", "Adam"],
            "batch_size": [5],
            "maximum_number_of_nodes": [-1],
            "maximum_number_of_features": [-1],
            "validation_split": [0.2],
            "test_split": [0.1, 0.2],
            "time_steps": [10],
            "validation_period": [5],
        }

        # When
        grid_search_dictionary = GridSearchParametersParser().get_grid_search_dictionary(
            epochs,
            loss_function_selection,
            optimizer_selection,
            batch_size,
            maximum_number_of_nodes,
            maximum_number_of_features,
            validation_split,
            test_split,
            time_steps,
            validation_period)

        # Then
        self.assertEqual(grid_search_dictionary_expected, grid_search_dictionary)



import logging

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.graph_encoder import GraphEncoder
from message_passing_nn.repository.file_system_repository import FileSystemRepository
from message_passing_nn.trainer.model_trainer import ModelTrainer
from message_passing_nn.usecase.grid_search import GridSearch
from message_passing_nn.utils.grid_search_parameters_parser import GridSearchParametersParser
from message_passing_nn.utils.saver import Saver


class MessagePassingNN:
    def __init__(self, grid_search: GridSearch) -> None:
        self.grid_search = grid_search

    def start(self):
        try:
            self.grid_search.start()
        except Exception:
            get_logger().exception("message")


def create(dataset_name: str,
           data_directory: str,
           model_directory: str,
           results_directory: str,
           device: str,
           epochs: str,
           loss_function_selection: str,
           optimizer_selection: str,
           batch_size: str,
           maximum_number_of_features: str,
           maximum_number_of_nodes: str,
           validation_split: str,
           test_split: str,
           time_steps: str,
           validation_period: str) -> MessagePassingNN:
    grid_search_dictionary = GridSearchParametersParser().get_grid_search_dictionary(epochs,
                                                                                     loss_function_selection,
                                                                                     optimizer_selection,
                                                                                     batch_size,
                                                                                     maximum_number_of_features,
                                                                                     maximum_number_of_nodes,
                                                                                     validation_split,
                                                                                     test_split,
                                                                                     time_steps,
                                                                                     validation_period)
    file_system_repository = FileSystemRepository(data_directory, dataset_name)
    data_preprocessor = DataPreprocessor()
    model_trainer = ModelTrainer(GraphEncoder, data_preprocessor, device)
    saver = Saver(model_directory, results_directory)
    grid_search = GridSearch(file_system_repository, data_preprocessor, model_trainer, grid_search_dictionary, saver)
    return MessagePassingNN(grid_search)


def get_logger() -> logging.Logger:
    return logging.getLogger('message_passing_nn')




from abc import ABCMeta, abstractmethod
from typing import Any, Tuple, List

import torch as to
from torch.utils.data import DataLoader


class Preprocessor(metaclass=ABCMeta):
    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def train_validation_test_split(dataset: Any,
                                    batch_size: int,
                                    maximum_number_of_nodes: int,
                                    maximum_number_of_features: int,
                                    validation_split: float = 0.2,
                                    test_split: float = 0.1) -> Tuple[DataLoader, DataLoader, DataLoader]:
        pass

    @staticmethod
    @abstractmethod
    def extract_data_dimensions(dataset: Any) -> Tuple:
        pass

    @staticmethod
    @abstractmethod
    def flatten(tensors: to.Tensor, desired_size: Any = 0) -> to.Tensor:
        pass

    @staticmethod
    @abstractmethod
    def normalize(tensors: to.Tensor) -> to.Tensor:
        pass



import logging
from typing import Dict, List, Tuple

import itertools
import numpy as np
from torch.utils.data.dataloader import DataLoader

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.repository.repository import Repository
from message_passing_nn.trainer.model_trainer import ModelTrainer
from message_passing_nn.utils.saver import Saver


class GridSearch:
    def __init__(self,
                 training_data_repository: Repository,
                 data_preprocessor: DataPreprocessor,
                 model_trainer: ModelTrainer,
                 grid_search_dictionary: Dict,
                 saver: Saver) -> None:
        self.repository = training_data_repository
        self.data_preprocessor = data_preprocessor
        self.model_trainer = model_trainer
        self.grid_search_dictionary = grid_search_dictionary
        self.saver = saver

    def start(self) -> Dict:
        all_grid_search_configurations = self._get_all_grid_search_configurations()
        self.get_logger().info('Started Training')
        losses = {'training_loss': {},
                  'validation_loss': {},
                  'test_loss': {}}
        configuration_id = ''
        for configuration in all_grid_search_configurations:
            configuration_id, grid_search_configuration_dictionary = self._get_grid_search_configuration_dictionary(
                configuration)
            losses = self._search_configuration(configuration_id,
                                                grid_search_configuration_dictionary,
                                                losses)
        self.saver.save_results(configuration_id, losses)
        self.get_logger().info('Finished Training')
        return losses

    def _search_configuration(self,
                              configuration_id: str,
                              grid_search_configuration_dictionary: Dict,
                              losses: Dict) -> Dict:
        training_data, validation_data, test_data, data_dimensions = self._prepare_dataset(
            grid_search_configuration_dictionary)
        self.model_trainer.instantiate_attributes(data_dimensions, grid_search_configuration_dictionary)
        losses = self._update_losses_with_configuration_id(grid_search_configuration_dictionary, losses)
        validation_loss_max = np.inf
        for epoch in range(1, grid_search_configuration_dictionary['epochs'] + 1):
            training_loss = self.model_trainer.do_train(training_data, epoch)
            losses['training_loss'][grid_search_configuration_dictionary["configuration_id"]].update(
                {epoch: training_loss})
            if epoch % grid_search_configuration_dictionary["validation_period"] == 0:
                validation_loss = self.model_trainer.do_evaluate(validation_data, epoch)
                losses['validation_loss'][grid_search_configuration_dictionary["configuration_id"]].update(
                    {epoch: validation_loss})
                if validation_loss < validation_loss_max:
                    self.saver.save_model(epoch, configuration_id, self.model_trainer.model)
        test_loss = self.model_trainer.do_evaluate(test_data)
        losses['test_loss'][grid_search_configuration_dictionary["configuration_id"]].update(
            {"final_epoch": test_loss})
        return losses

    @staticmethod
    def _update_losses_with_configuration_id(configuration_dictionary: Dict, losses: Dict) -> Dict:
        losses['training_loss'].update({configuration_dictionary["configuration_id"]: {}})
        losses['validation_loss'].update({configuration_dictionary["configuration_id"]: {}})
        losses['test_loss'].update({configuration_dictionary["configuration_id"]: {}})
        return losses

    @staticmethod
    def _get_grid_search_configuration_dictionary(configuration: Tuple[Tuple]) -> Tuple[str, Dict]:
        grid_search_configuration_dictionary = dict(((key, value) for key, value in configuration))
        configuration_id = 'configuration_id'
        for key, value in grid_search_configuration_dictionary.items():
            configuration_id += "__" + "_".join([key, str(value)])
        grid_search_configuration_dictionary.update({"configuration_id": configuration_id})
        return configuration_id, grid_search_configuration_dictionary

    def _prepare_dataset(self, configuration_dictionary: Dict) -> Tuple[DataLoader, DataLoader, DataLoader, Tuple]:
        raw_dataset = self.repository.get_all_data()
        equalized_dataset = self.data_preprocessor.equalize_dataset_dimensions(raw_dataset,
                                                                               configuration_dictionary['maximum_number_of_nodes'],
                                                                               configuration_dictionary['maximum_number_of_features'])
        training_data, validation_data, test_data = self.data_preprocessor \
            .train_validation_test_split(equalized_dataset,
                                         configuration_dictionary['batch_size'],
                                         configuration_dictionary['maximum_number_of_nodes'],
                                         configuration_dictionary['maximum_number_of_features'],
                                         configuration_dictionary['validation_split'],
                                         configuration_dictionary['test_split'])
        data_dimensions = self.data_preprocessor.extract_data_dimensions(raw_dataset)
        return training_data, validation_data, test_data, data_dimensions

    def _get_all_grid_search_configurations(self) -> List[Tuple[Tuple]]:
        all_grid_search_configurations = []
        for key in self.grid_search_dictionary.keys():
            all_grid_search_configurations.append([(key, value) for value in self.grid_search_dictionary[key]])
        return list(itertools.product(*all_grid_search_configurations))

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



import math

import torch as to
import torch.nn as nn
from typing import List

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.node import Node


class GraphEncoder(nn.Module):
    def __init__(self,
                 time_steps: int,
                 number_of_nodes: int,
                 number_of_node_features: int,
                 fully_connected_layer_input_size: int,
                 fully_connected_layer_output_size: int,
                 device: str) -> None:
        super(GraphEncoder, self).__init__()
        base_4d_tensor_shape = [number_of_nodes, number_of_nodes, number_of_node_features, number_of_node_features]
        base_3d_tensor_shape = [number_of_nodes, number_of_node_features, number_of_node_features]
        base_2d_tensor_shape = [number_of_node_features, 1]

        self.time_steps = time_steps
        self.number_of_nodes = number_of_nodes
        self.number_of_node_features = number_of_node_features
        self.fully_connected_layer_input_size = fully_connected_layer_input_size
        self.fully_connected_layer_output_size = fully_connected_layer_output_size
        self.device = device

        self.w_gru_update_gate_features = self._get_parameter(base_4d_tensor_shape)
        self.w_gru_forget_gate_features = self._get_parameter(base_4d_tensor_shape)
        self.w_gru_current_memory_message_features = self._get_parameter(base_4d_tensor_shape)
        self.u_gru_update_gate = self._get_parameter(base_4d_tensor_shape)
        self.u_gru_forget_gate = self._get_parameter(base_4d_tensor_shape)
        self.u_gru_current_memory_message = self._get_parameter(base_4d_tensor_shape)
        self.b_gru_update_gate = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
        self.b_gru_forget_gate = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
        self.b_gru_current_memory_message = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
        self.u_graph_node_features = self._get_parameter(base_3d_tensor_shape)
        self.u_graph_neighbor_messages = self._get_parameter(base_3d_tensor_shape)
        self.linear = to.nn.Linear(self.fully_connected_layer_input_size, self.fully_connected_layer_output_size)
        self.sigmoid = to.nn.Sigmoid()

    @classmethod
    def of(cls,
           time_steps: int,
           number_of_nodes: int,
           number_of_node_features: int,
           fully_connected_layer_input_size: int,
           fully_connected_layer_output_size: int,
           device: str):
        return cls(time_steps,
                   number_of_nodes,
                   number_of_node_features,
                   fully_connected_layer_input_size,
                   fully_connected_layer_output_size,
                   device)

    def forward(self,
                node_features: to.Tensor,
                adjacency_matrix: to.Tensor,
                batch_size: int) -> to.Tensor:
        outputs = to.zeros(batch_size, self.fully_connected_layer_output_size, device=self.device)
        for batch in range(batch_size):
            outputs[batch] = self.sigmoid(
                self.linear(
                    DataPreprocessor.flatten(
                        self.encode(node_features[batch], adjacency_matrix[batch]),
                        self.fully_connected_layer_input_size)))
        return outputs

    def encode(self, node_features: to.Tensor, adjacency_matrix: to.Tensor) -> to.Tensor:
        messages = self._send_messages(node_features, adjacency_matrix)
        encodings = self._encode_nodes(node_features, messages)
        return encodings

    def _send_messages(self, node_features: to.Tensor, adjacency_matrix: to.Tensor) -> to.Tensor:
        messages = to.zeros((self.number_of_nodes, self.number_of_nodes, self.number_of_node_features),
                            device=self.device)
        for step in range(self.time_steps):
            messages = self._compose_messages(node_features, adjacency_matrix, messages)
        return messages

    def _encode_nodes(self, node_features: to.Tensor, messages: to.Tensor) -> to.Tensor:
        encoded_node = to.zeros(self.number_of_nodes, self.number_of_node_features, device=self.device)
        for node_id in range(self.number_of_nodes):
            encoded_node[node_id] = self._apply_recurrent_layer(node_features, messages, node_id)
        return encoded_node

    def _apply_recurrent_layer(self, node_features: to.Tensor, messages: to.Tensor, node_id: int) -> to.Tensor:
        node_encoding_features = self.u_graph_node_features[node_id].matmul(node_features[node_id])
        node_encoding_messages = self.u_graph_neighbor_messages[node_id].matmul(to.sum(messages[node_id], dim=0))
        return to.relu(node_encoding_features + node_encoding_messages)

    def _compose_messages(self,
                          node_features: to.Tensor,
                          adjacency_matrix: to.Tensor,
                          messages: to.Tensor) -> to.Tensor:
        new_messages = to.zeros(messages.shape, device=self.device)
        for node_id in range(self.number_of_nodes):
            node = self._create_node(node_features, adjacency_matrix, node_id)
            for end_node_id in node.neighbors:
                message = self._get_message_value(messages, node, end_node_id, node_features)
                new_messages[node_id, end_node_id] = message
                if Node.is_adjacency_matrix_symmetric(adjacency_matrix):
                    new_messages[end_node_id, node_id] = message
        return new_messages

    def _get_message_value(self,
                           messages: to.Tensor,
                           node: Node,
                           end_node_id: int,
                           node_features: to.Tensor) -> to.Tensor:
        previous_messages = self._sum_messages_from_neighbors_except_target(messages,
                                                                            node,
                                                                            end_node_id)
        update_gate = self._pass_through_update_gate(messages, node, end_node_id, node_features)
        current_memory = self._get_current_memory_message(messages, node, end_node_id, node_features)
        return to.add(
            to.mul(
                to.sub(to.ones(update_gate.shape, device=self.device),
                       update_gate),
                previous_messages),
            to.mul(update_gate,
                   current_memory))

    def _sum_messages_from_neighbors_except_target(self,
                                                   messages: to.Tensor,
                                                   node: Node,
                                                   end_node_id: int) -> to.Tensor:
        messages_from_the_other_neighbors = to.zeros(node.features.shape[0], device=self.device)
        if node.neighbors_count > 1:
            neighbors_slice = node.get_start_node_neighbors_without_end_node(end_node_id)
            messages_from_the_other_neighbors = to.sum(messages[neighbors_slice], dim=0)
        return messages_from_the_other_neighbors

    def _pass_through_update_gate(self,
                                  messages: to.Tensor,
                                  node: Node,
                                  end_node_id: int,
                                  node_features: to.Tensor) -> to.Tensor:
        message_from_a_neighbor_other_than_target = self._sum_messages_from_neighbors_except_target(
            messages,
            node,
            end_node_id)
        update_gate_output = to.sigmoid(
            to.add(
                to.add(self.w_gru_update_gate_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(
                           message_from_a_neighbor_other_than_target)),
                self.b_gru_update_gate))
        return update_gate_output

    def _get_current_memory_message(self,
                                    messages: to.Tensor,
                                    node: Node,
                                    end_node_id: int,
                                    node_features: to.Tensor) -> to.Tensor:
        messages_passed_through_reset_gate = self._keep_or_reset_messages(messages,
                                                                          node,
                                                                          end_node_id,
                                                                          node_features)
        current_memory_message = to.add(
            to.add(self.w_gru_current_memory_message_features[node.node_id, end_node_id].matmul(
                node_features[node.node_id]),
                   self.u_gru_current_memory_message[node.node_id, end_node_id].matmul(
                       messages_passed_through_reset_gate)),
            self.b_gru_current_memory_message)
        return to.tanh(current_memory_message)

    def _keep_or_reset_messages(self,
                                messages: to.Tensor,
                                node: Node,
                                end_node_id: int,
                                node_features: to.Tensor) -> to.Tensor:
        return self.u_gru_current_memory_message[node.node_id, end_node_id].matmul(sum([to.mul(to.sigmoid(
            to.add(
                to.add(self.w_gru_update_gate_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(
                           messages[node.node_id, reset_node_index])),
                self.b_gru_update_gate)).long(), messages[node.node_id, reset_node_index])
                                                                                        for reset_node_index in
                                                                                        node.get_start_node_neighbors_without_end_node(
                                                                                            end_node_id)[0]]))

    def _pass_through_reset_gate(self,
                                 messages: to.Tensor,
                                 node: Node,
                                 end_node_id: int,
                                 node_features: to.Tensor) -> to.Tensor:
        return to.sigmoid(
            to.add(
                to.add(self.w_gru_update_gate_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(messages[node.node_id, end_node_id])),
                self.b_gru_update_gate)).long()

    def _get_parameter(self, tensor_shape: List[int]) -> nn.Parameter:
        return nn.Parameter(nn.init.kaiming_normal_(to.zeros(tensor_shape, device=self.device)), requires_grad=True)

    @staticmethod
    def _create_node(node_features: to.Tensor, adjacency_matrix: to.Tensor, node_id: int) -> Node:
        return Node(node_features, adjacency_matrix, node_id)



from message_passing_nn.utils.grid_search_parameters_parser import GridSearchParametersParser
from message_passing_nn.utils.loss_function_selector import LossFunctionSelector
from message_passing_nn.utils.optimizer_selector import OptimizerSelector
from message_passing_nn.utils.saver import Saver



from unittest import TestCase

import numpy as np

from message_passing_nn.model.node import Node
from tests.fixtures.matrices_and_vectors import BASE_GRAPH_NODE_FEATURES, BASE_GRAPH, NOT_SYMMETRIC_GRAPH


class TestNode(TestCase):
    def setUp(self) -> None:
        all_node_features = BASE_GRAPH_NODE_FEATURES
        labels = BASE_GRAPH
        self.node = Node(all_node_features, labels, node_id=2)

    def test_is_matrix_symmetric(self):
        # Given
        symmetric_matrix = BASE_GRAPH

        # When
        symmetry = Node.is_adjacency_matrix_symmetric(symmetric_matrix)

        # Then
        self.assertTrue(symmetry)

    def test_is_matrix_not_symmetric(self):
        # Given
        not_symmetric_matrix = NOT_SYMMETRIC_GRAPH

        # When
        symmetry = Node.is_adjacency_matrix_symmetric(not_symmetric_matrix)

        # Then
        self.assertFalse(symmetry)

    def test_node_neighbors_symmetric(self):
        # Given
        neighbors_expected = np.array([3])

        # When
        neighbors = self.node.neighbors

        # Then
        self.assertTrue(np.array_equal(neighbors_expected, neighbors))

    def test_node_neighbors_not_symmetric(self):
        # Given
        all_node_features = BASE_GRAPH_NODE_FEATURES
        labels = NOT_SYMMETRIC_GRAPH
        self.node = Node(all_node_features, labels, node_id=2)
        neighbors_expected = np.array([0, 1, 3])

        # When
        neighbors = self.node.neighbors

        # Then
        self.assertTrue(np.array_equal(neighbors_expected, neighbors))

    def test_get_start_node_neighbors_without_end_node(self):
        # Given
        start_node_neighbors_expected = np.array([1, 3])
        end_node_id = 0

        # When
        start_node_neighbors = self.node.get_start_node_neighbors_without_end_node(end_node_id)[0]

        # Then
        self.assertTrue(np.array_equal(start_node_neighbors_expected, start_node_neighbors))



from unittest import TestCase

import torch as to
from torch import nn

from message_passing_nn.model.graph_encoder import GraphEncoder
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES, \
    BASE_U_MATRIX, BASE_W_MATRIX, MULTIPLICATION_FACTOR, BASE_B_VECTOR


class TestGraphEncoder(TestCase):

    def setUp(self) -> None:
        number_of_nodes = BASE_GRAPH.size()[0]
        number_of_node_features = BASE_GRAPH_NODE_FEATURES.size()[1]
        device = "cpu"
        self.graph_encoder = GraphEncoder.of(time_steps=1,
                                             number_of_nodes=number_of_nodes,
                                             number_of_node_features=number_of_node_features,
                                             fully_connected_layer_input_size=number_of_nodes * number_of_node_features,
                                             fully_connected_layer_output_size=number_of_nodes ** 2,
                                             device=device)
        self.graph_encoder.w_gru_update_gate_features = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                                     requires_grad=False).float()
        self.graph_encoder.w_gru_forget_gate_features = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                                     requires_grad=False).float()
        self.graph_encoder.w_gru_current_memory_message_features = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                                                requires_grad=False).float()
        self.graph_encoder.u_gru_update_gate = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                            requires_grad=False).float()
        self.graph_encoder.u_gru_forget_gate = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                            requires_grad=False).float()
        self.graph_encoder.u_gru_current_memory_message = nn.Parameter(MULTIPLICATION_FACTOR * BASE_W_MATRIX,
                                                                       requires_grad=False).float()
        self.graph_encoder.b_gru_update_gate = nn.Parameter(MULTIPLICATION_FACTOR * BASE_B_VECTOR,
                                                            requires_grad=False).float()
        self.graph_encoder.b_gru_forget_gate = nn.Parameter(MULTIPLICATION_FACTOR * BASE_B_VECTOR,
                                                            requires_grad=False).float()
        self.graph_encoder.b_gru_current_memory_message = nn.Parameter(MULTIPLICATION_FACTOR * BASE_B_VECTOR,
                                                                       requires_grad=False).float()
        self.graph_encoder.u_graph_node_features = nn.Parameter(MULTIPLICATION_FACTOR * BASE_U_MATRIX,
                                                                requires_grad=False).float()
        self.graph_encoder.u_graph_neighbor_messages = nn.Parameter(MULTIPLICATION_FACTOR * BASE_U_MATRIX,
                                                                    requires_grad=False).float()
        self.graph_encoder.linear.weight = to.nn.Parameter(to.ones(number_of_nodes ** 2,
                                                                   number_of_nodes * number_of_node_features),
                                                           requires_grad=False).float()
        self.graph_encoder.linear.bias = to.nn.Parameter(2 * to.ones(number_of_nodes ** 2), requires_grad=False).float()

    def test_forward(self):
        # Given
        batch_size = 1
        features = BASE_GRAPH_NODE_FEATURES.view(1, BASE_GRAPH.size()[0], BASE_GRAPH_NODE_FEATURES.size()[1])
        adjacency_matrix = BASE_GRAPH.view(1, BASE_GRAPH.size()[0], BASE_GRAPH.size()[1],)
        outputs_expected = to.tensor([[0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733,
                                       0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733, 0.98733]])

        # When
        outputs = self.graph_encoder.forward(features, adjacency_matrix, batch_size)

        # Then
        self.assertTrue(to.allclose(outputs_expected, outputs))

    def test_encode_graph_returns_the_expected_encoding_for_a_node_after_one_time_step(self):
        # Give
        node = 0
        node_encoding_expected = to.tensor([[0.3909883, 0.3909883]])

        # When
        node_encoding = self.graph_encoder.encode(BASE_GRAPH_NODE_FEATURES, BASE_GRAPH)[node]

        # Then
        self.assertTrue(to.allclose(node_encoding_expected, node_encoding))

    def test_encode_graph_returns_the_expected_shape(self):
        # Given
        encoded_graph_shape_expected = list(BASE_GRAPH_NODE_FEATURES.shape)

        # When
        encoded_graph_shape = self.graph_encoder.encode(BASE_GRAPH_NODE_FEATURES, BASE_GRAPH).shape

        # Then
        self.assertEqual(encoded_graph_shape_expected, list(encoded_graph_shape))



import os.path
from os import path
from unittest import TestCase

import torch as to

from message_passing_nn.repository.file_system_repository import FileSystemRepository
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestTrainingDataRepository(TestCase):
    def setUp(self) -> None:
        self.dataset = 'repo-test-data'
        self.tests_data_directory = 'tests/test_data/'
        self.file_system_repository = FileSystemRepository(self.tests_data_directory, self.dataset)

    def test_save(self):
        # Given
        features = BASE_GRAPH_NODE_FEATURES
        adjacency_matrix = BASE_GRAPH
        labels = BASE_GRAPH.view(-1)

        filenames_to_save = ['code_features.pickle', 'code_adjacency-matrix.pickle', 'code_labels.pickle']
        filenames_expected = [self.tests_data_directory + self.dataset + '/code_features.pickle',
                              self.tests_data_directory + self.dataset + '/code_adjacency-matrix.pickle',
                              self.tests_data_directory + self.dataset + '/code_labels.pickle']

        # When
        self.file_system_repository.save(filenames_to_save[0], features)
        self.file_system_repository.save(filenames_to_save[1], adjacency_matrix)
        self.file_system_repository.save(filenames_to_save[2], labels)

        # Then
        path.exists(filenames_expected[0])
        path.exists(filenames_expected[1])
        path.exists(filenames_expected[2])
        os.remove(filenames_expected[0])
        os.remove(filenames_expected[1])
        os.remove(filenames_expected[2])

    def test_get_all_data(self):
        # Given
        features_expected = BASE_GRAPH_NODE_FEATURES
        adjacency_matrix_expected = BASE_GRAPH
        labels_expected = BASE_GRAPH.view(-1)
        filenames_to_save = ['code_features.pickle', 'code_adjacency-matrix.pickle', 'code_labels.pickle']
        self.file_system_repository.save(filenames_to_save[0], features_expected)
        self.file_system_repository.save(filenames_to_save[1], adjacency_matrix_expected)
        self.file_system_repository.save(filenames_to_save[2], labels_expected)
        all_data_expected = [(features_expected, adjacency_matrix_expected, labels_expected)]

        # When
        all_data = self.file_system_repository.get_all_data()

        # Then
        self.assertTrue(to.allclose(all_data_expected[0][0], all_data[0][0]))
        self.assertTrue(to.allclose(all_data_expected[0][1], all_data[0][1]))
        self.assertTrue(to.allclose(all_data_expected[0][2], all_data[0][2]))
